# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""A torch function library for onnxscript.

The modules are named after the torch module names for grouping:

- core
- nn
- fft
- vision
- etc.
"""
